from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from typing import List, Dict, Tuple
import torch
from relogic.logickit.utils import create_tensor

from relogic.logickit.tokenizer.tokenization import BertTokenizer

class DependencyParsingExample(Example):
  """DependencyParsingExample


  """
  def __init__(self, text, arcs=None, labels=None, lang=None):
    super(DependencyParsingExample, self).__init__()
    self.text = text
    self.raw_tokens = text.split()
    self.arcs = arcs
    self.labels = labels
    self.lang = lang
    self.label_padding = "X"

  def process(self, tokenizers: Dict, *inputs, **kwargs):
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, BertTokenizer):
        self.text_tokens, self.text_is_head = tokenizer.tokenize(self.text)

        self.tokens = ["[CLS]"] + self.text_tokens + ["[SEP]"]
        self.segment_ids = [0] * (len(self.tokens))
        self.is_head = [2] + self.text_is_head + [2]
        self.head_index = [idx for idx, value in enumerate(self.is_head) if value == 1]

        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)

        if self.arcs is not None and self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          self.label_ids = [self.label_padding_id] * len(self.input_ids)
          self.arcs_ids = [-1] * len(self.input_ids)
          assert(len(self.labels) == len(self.arcs))
          for idx, label, arc in zip(self.head_index, self.labels, self.arcs):
            self.label_ids[idx] = label_mapping[label]
            self.arcs_ids[idx] = arc
        else:
          self.label_ids = None
          self.arcs_ids = None

        if self.lang is not None:
          language_name2id = kwargs.get("language_name2id")
          if language_name2id is not None:
            self.lang_id = language_name2id[self.lang]

  @classmethod
  def from_structure(cls, structure):
    return cls(text=structure.text)

  @classmethod
  def from_json(cls, example):
    return cls(text=" ".join(example["tokens"]),
               arcs=example.get("arcs", None),
               labels=example.get("labels", None),
               lang=example.get("lang", None))

  @property
  def len(self):
      return len(self.input_ids)


class DependencyParsingFeature(Feature):
  """
  Sequence Features
  """
  def __init__(self, *inputs, **kwargs):
    super().__init__()
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.is_head = kwargs.pop("is_head")
    self.arcs_ids = kwargs.pop("arcs_ids")
    self.label_ids = kwargs.pop("label_ids")
    self.lang_ids = kwargs.pop("lang_ids", None)

class DependencyParsingMiniBatch(MiniBatch):
  def __init__(self, *inputs, **kwargs):
    super().__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids", torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask", torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids", torch.long, device)
    inputs["input_head"] = create_tensor(self.input_features, "is_head", torch.long, device)

    if use_label:
      label_ids = create_tensor(self.input_features, "label_ids", torch.long, device)
      inputs["label_ids"] = label_ids
      arcs_ids = create_tensor(self.input_features, "arcs_ids", torch.long, device)
      inputs["arcs_ids"] = arcs_ids
    else:
      inputs["label_ids"] = None
      inputs["arcs_ids"] = None

    inputs["lang_ids"] = create_tensor(self.input_features, "lang_ids",
                                       torch.long, device)
    inputs["extra_args"] = {}
    return inputs

class DependencyParsingDataFlow(DataFlow):
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super().__init__(config, task_name, tokenizers, label_mapping)
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}

  @property
  def example_class(self):
    return DependencyParsingExample

  @property
  def minibatch_class(self):
    return DependencyParsingMiniBatch

  def process_example(self, example: DependencyParsingExample):
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping,
                    language_name2id=None)

  def convert_examples_to_features(self, examples: List[DependencyParsingExample]):
    examples: List[DependencyParsingExample]
    features = []

    max_token_length = max([example.len for example in examples])

    for idx, example in enumerate(examples):
      padding = [0] * (max_token_length - example.len)
      input_ids = example.input_ids + padding
      input_mask = example.input_mask + padding
      segment_ids = example.segment_ids + padding
      is_head = example.is_head + [2] * (max_token_length - example.len)

      if example.label_ids is not None and example.arcs_ids is not None:
        label_ids = example.label_ids + [example.label_padding_id] * (max_token_length - example.len)
        arcs_ids = example.arcs_ids + [-1] * (max_token_length - example.len)
      else:
        label_ids = None
        arcs_ids = None

      features.append(DependencyParsingFeature(
        input_ids=input_ids,
        input_mask=input_mask,
        is_head=is_head,
        segment_ids=segment_ids,
        label_ids=label_ids,
        arcs_ids=arcs_ids))

    return features


        


import torch

from typing import List
import itertools

def block_orthogonal(tensor: torch.Tensor,
                     split_sizes: List[int],
                     gain: float = 1.0) -> None:
  """
  An initializer which allows initializing model parameters in "blocks". This is helpful
  in the case of recurrent models which use multiple gates applied to linear projections,
  which can be computed efficiently if they are concatenated together. However, they are
  separate parameters which should be initialized independently.
  Parameters
  ----------
  tensor : ``torch.Tensor``, required.
      A tensor to initialize.
  split_sizes : List[int], required.
      A list of length ``tensor.ndim()`` specifying the size of the
      blocks along that particular dimension. E.g. ``[10, 20]`` would
      result in the tensor being split into chunks of size 10 along the
      first dimension and 20 along the second.
  gain : float, optional (default = 1.0)
      The gain (scaling) applied to the orthogonal initialization.
  """
  data = tensor.data
  sizes = list(tensor.size())
  if any([a % b != 0 for a, b in zip(sizes, split_sizes)]):
    raise ValueError("tensor dimensions must be divisible by their respective "
                             "split_sizes. Found size: {} and split_sizes: {}".format(sizes, split_sizes))
  indexes = [list(range(0, max_size, split))
             for max_size, split in zip(sizes, split_sizes)]
  # Iterate over all possible blocks within the tensor.
  for block_start_indices in itertools.product(*indexes):
    # A list of tuples containing the index to start at for this block
    # and the appropriate step size (i.e split_size[i] for dimension i).
    index_and_step_tuples = zip(block_start_indices, split_sizes)
    # This is a tuple of slices corresponding to:
    # tensor[index: index + step_size, ...]. This is
    # required because we could have an arbitrary number
    # of dimensions. The actual slices we need are the
    # start_index: start_index + step for each dimension in the tensor.
    block_slice = tuple([slice(start_index, start_index + step)
                         for start_index, step in index_and_step_tuples])
    data[block_slice] = torch.nn.init.orthogonal_(tensor[block_slice].contiguous(), gain=gain)


# coding=utf-8
# Copyright 2019-present, Facebook, Inc and the HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" XLM configuration """
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
import sys
from io import open

from relogic.logickit.base.configuration_utils import PretrainedConfig

logger = logging.getLogger(__name__)

XLM_PRETRAINED_CONFIG_ARCHIVE_MAP = {
    'xlm-mlm-en-2048': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-en-2048-config.json",
    'xlm-mlm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-ende-1024-config.json",
    'xlm-mlm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enfr-1024-config.json",
    'xlm-mlm-enro-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enro-1024-config.json",
    'xlm-mlm-tlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-tlm-xnli15-1024-config.json",
    'xlm-mlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-xnli15-1024-config.json",
    'xlm-clm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-clm-enfr-1024-config.json",
    'xlm-clm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-clm-ende-1024-config.json",
    'xlm-mlm-17-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-17-1280-config.json",
    'xlm-mlm-100-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-100-1280-config.json",
}


class XLMConfig(PretrainedConfig):
    """Configuration class to store the configuration of a `XLMModel`.

    Args:
        vocab_size_or_config_json_file: Vocabulary size of `inputs_ids` in `XLMModel`.
        d_model: Size of the encoder layers and the pooler layer.
        n_layer: Number of hidden layers in the Transformer encoder.
        n_head: Number of attention heads for each attention layer in
            the Transformer encoder.
        d_inner: The size of the "intermediate" (i.e., feed-forward)
            layer in the Transformer encoder.
        ff_activation: The non-linear activation function (function or string) in the
            encoder and pooler. If string, "gelu", "relu" and "swish" are supported.
        untie_r: untie relative position biases
        attn_type: 'bi' for XLM, 'uni' for Transformer-XL

        dropout: The dropout probabilitiy for all fully connected
            layers in the embeddings, encoder, and pooler.
        max_position_embeddings: The maximum sequence length that this model might
            ever be used with. Typically set this to something large just in case
            (e.g., 512 or 1024 or 2048).
        initializer_range: The sttdev of the truncated_normal_initializer for
            initializing all weight matrices.
        layer_norm_eps: The epsilon used by LayerNorm.

        dropout: float, dropout rate.
        init: str, the initialization scheme, either "normal" or "uniform".
        init_range: float, initialize the parameters with a uniform distribution
            in [-init_range, init_range]. Only effective when init="uniform".
        init_std: float, initialize the parameters with a normal distribution
            with mean 0 and stddev init_std. Only effective when init="normal".
        mem_len: int, the number of tokens to cache.
        reuse_len: int, the number of tokens in the currect batch to be cached
            and reused in the future.
        bi_data: bool, whether to use bidirectional input pipeline.
            Usually set to True during pretraining and False during finetuning.
        clamp_len: int, clamp all relative distances larger than clamp_len.
            -1 means no clamping.
        same_length: bool, whether to use the same attention length for each token.
    """
    pretrained_config_archive_map = XLM_PRETRAINED_CONFIG_ARCHIVE_MAP

    def __init__(self,
                 vocab_size_or_config_json_file=30145,
                 emb_dim=2048,
                 n_layers=12,
                 n_heads=16,
                 dropout=0.1,
                 attention_dropout=0.1,
                 gelu_activation=True,
                 sinusoidal_embeddings=False,
                 causal=False,
                 asm=False,
                 n_langs=1,
                 use_lang_emb=True,
                 max_position_embeddings=512,
                 embed_init_std=2048 ** -0.5,
                 layer_norm_eps=1e-12,
                 init_std=0.02,
                 bos_index=0,
                 eos_index=1,
                 pad_index=2,
                 unk_index=3,
                 mask_index=5,
                 is_encoder=True,

                 finetuning_task=None,
                 num_labels=2,
                 summary_type='first',
                 summary_use_proj=True,
                 summary_activation=None,
                 summary_proj_to_labels=True,
                 summary_first_dropout=0.1,
                 start_n_top=5,
                 end_n_top=5,
                 **kwargs):
        """Constructs XLMConfig.
        """
        super(XLMConfig, self).__init__(**kwargs)

        if isinstance(vocab_size_or_config_json_file, str) or (sys.version_info[0] == 2
                        and isinstance(vocab_size_or_config_json_file, unicode)):
            with open(vocab_size_or_config_json_file, "r", encoding='utf-8') as reader:
                json_config = json.loads(reader.read())
            for key, value in json_config.items():
                self.__dict__[key] = value
        elif isinstance(vocab_size_or_config_json_file, int):
            self.n_words = vocab_size_or_config_json_file
            self.emb_dim = emb_dim
            self.n_layers = n_layers
            self.n_heads = n_heads
            self.dropout = dropout
            self.attention_dropout = attention_dropout
            self.gelu_activation = gelu_activation
            self.sinusoidal_embeddings = sinusoidal_embeddings
            self.causal = causal
            self.asm = asm
            self.n_langs = n_langs
            self.use_lang_emb = use_lang_emb
            self.layer_norm_eps = layer_norm_eps
            self.bos_index = bos_index
            self.eos_index = eos_index
            self.pad_index = pad_index
            self.unk_index = unk_index
            self.mask_index = mask_index
            self.is_encoder = is_encoder
            self.max_position_embeddings = max_position_embeddings
            self.embed_init_std = embed_init_std
            self.init_std = init_std
            self.finetuning_task = finetuning_task
            self.num_labels = num_labels
            self.summary_type = summary_type
            self.summary_use_proj = summary_use_proj
            self.summary_activation = summary_activation
            self.summary_proj_to_labels = summary_proj_to_labels
            self.summary_first_dropout = summary_first_dropout
            self.start_n_top = start_n_top
            self.end_n_top = end_n_top
        else:
            raise ValueError("First argument must be either a vocabulary size (int)"
                             " or the path to a pretrained model config file (str)")

    @property
    def vocab_size(self):
        return self.n_words

    @vocab_size.setter
    def vocab_size(self, value):
        self.n_words = value

    @property
    def hidden_size(self):
        return self.emb_dim

    @property
    def num_attention_heads(self):
        return self.n_heads

    @property
    def num_hidden_layers(self):
        return self.n_layers



import tensorflow as tf
import os
import time

from relogic.tpukit.model.classifier_model import model_fn_builder
from relogic.tpukit.inference import modeling
from relogic.tpukit.processors import processors
from relogic.tpukit.input_fn import file_based_input_fn_builder, file_based_convert_examples_to_features, PaddingInputExample
from relogic.tpukit.tokenizer import tokenization

class Trainer(object):
  def __init__(self, FLAGS):
    self.FLAGS = FLAGS
    self.bert_config = modeling.BertConfig.from_json_file(FLAGS.bert_config_file)
    self.task_name = FLAGS.task_name.lower()
    self.tokenizer = tokenization.FullTokenizer(
      vocab_file=FLAGS.vocab_file, do_lower_case=FLAGS.do_lower_case)

    processor = processors[self.task_name]()
    self.label_list = processor.get_labels()

    tpu_cluster_resolver = None
    if FLAGS.use_tpu and FLAGS.tpu_name:
      tpu_cluster_resolver = tf.contrib.cluster_resolver.TPUClusterResolver(
        FLAGS.tpu_name, zone=FLAGS.tpu_zone, project=FLAGS.gcp_project)

    is_per_host = tf.contrib.tpu.InputPipelineConfig.PER_HOST_V2
    run_config = tf.contrib.tpu.RunConfig(
      cluster=tpu_cluster_resolver,
      master=FLAGS.master,
      model_dir=FLAGS.output_dir,
      save_checkpoints_steps=FLAGS.save_checkpoints_steps,
      tpu_config=tf.contrib.tpu.TPUConfig(
        iterations_per_loop=FLAGS.iterations_per_loop,
        num_shards=FLAGS.num_tpu_cores,
        per_host_input_for_training=is_per_host))

    if not FLAGS.do_pretraining and FLAGS.do_train:

      # Setup Training
      self.train_examples = None
      num_train_steps = None
      num_warmup_steps = None
      if FLAGS.do_train:
        self.train_examples = processor.get_train_examples(FLAGS.data_dir)
        num_train_steps = int(
          len(self.train_examples) / FLAGS.train_batch_size * FLAGS.num_train_epochs)
        num_warmup_steps = int(num_train_steps * FLAGS.warmup_proportion)
        train_file = os.path.join(self.FLAGS.output_dir, "train.tf_record")
        file_based_convert_examples_to_features(
          self.train_examples, self.label_list, self.FLAGS.max_seq_length, self.tokenizer, train_file)
        tf.logging.info("***** Running training *****")
        tf.logging.info("  Num examples = %d", len(self.train_examples))
        tf.logging.info("  Batch size = %d", self.FLAGS.train_batch_size)
        tf.logging.info("  Num steps = %d", num_train_steps)
        self.train_input_fn = file_based_input_fn_builder(
          input_file=train_file,
          seq_length=self.FLAGS.max_seq_length,
          is_training=True,
          drop_remainder=True)

        # Setup Validation
        self.dev_examples = processor.get_dev_examples(FLAGS.data_dir)
        num_actual_eval_examples = len(self.dev_examples)
        if FLAGS.use_tpu:
          # TPU requires a fixed batch size for all batches, therefore the number
          # of examples must be a multiple of the batch size, or else examples
          # will get dropped. So we pad with fake examples which are ignored
          # later on. These do NOT count towards the metric (all tf.metrics
          # support a per-instance weight, and these get a weight of 0.0).
          while len(self.dev_examples) % FLAGS.eval_batch_size != 0:
            self.dev_examples.append(PaddingInputExample())
        eval_file = os.path.join(FLAGS.output_dir, "eval.tf_record")
        file_based_convert_examples_to_features(
          self.dev_examples, self.label_list, FLAGS.max_seq_length, self.tokenizer, eval_file)

        tf.logging.info("***** Running evaluation *****")
        tf.logging.info("  Num examples = %d (%d actual, %d padding)",
                        len(self.dev_examples), num_actual_eval_examples,
                        len(self.dev_examples) - num_actual_eval_examples)
        tf.logging.info("  Batch size = %d", FLAGS.eval_batch_size)

        # This tells the estimator to run through the entire set.
        self.eval_steps = None
        # However, if running eval on the TPU, you will need to specify the
        # number of steps.
        if FLAGS.use_tpu:
          assert len(self.dev_examples) % FLAGS.eval_batch_size == 0
          self.eval_steps = int(len(self.dev_examples) // FLAGS.eval_batch_size)

        eval_drop_remainder = True if FLAGS.use_tpu else False
        self.eval_input_fn = file_based_input_fn_builder(
          input_file=eval_file,
          seq_length=FLAGS.max_seq_length,
          is_training=False,
          drop_remainder=eval_drop_remainder)

    if FLAGS.do_pretraining:
      num_train_steps = FLAGS.num_train_steps
      num_warmup_steps = FLAGS.num_warmup_steps

    self.num_train_steps = num_train_steps

    model_fn = model_fn_builder(
      bert_config=self.bert_config,
      num_labels=len(self.label_list),
      init_checkpoint=FLAGS.init_checkpoint,
      learning_rate=FLAGS.learning_rate,
      num_train_steps=num_train_steps,
      num_warmup_steps=num_warmup_steps,
      use_tpu=FLAGS.use_tpu,
      use_one_hot_embeddings=FLAGS.use_tpu)

    self.estimator = tf.contrib.tpu.TPUEstimator(
      use_tpu=FLAGS.use_tpu,
      model_fn=model_fn,
      config=run_config,
      train_batch_size=FLAGS.train_batch_size,
      eval_batch_size=FLAGS.eval_batch_size,
      predict_batch_size=FLAGS.predict_batch_size)

  def train(self):
    # heading = lambda s: utils.heading(s, '(' + self.config.model_name + ')')
    current_step = 0

    if not self.FLAGS.do_pretraining:

      start_timestamp = time.time()
      # while current_step < self.num_train_steps:
      #   next_checkpoint = min(current_step + self.FLAGS.eval_dev_every, self.num_train_steps)
      self.estimator.train(input_fn=self.train_input_fn, max_steps=self.num_train_steps)
      # current_step = next_checkpoint
      tf.logging.info('Finished training up to step %d. Elapsed seconds %d.',
                      self.num_train_steps, int(time.time() - start_timestamp))
      tf.logging.info('Starting to evaluate at step %d', self.num_train_steps)

      self.evaluate_task()

    else:
      raise NotImplementedError()

  def evaluate_task(self):


    result = self.estimator.evaluate(input_fn=self.eval_input_fn, steps=self.eval_steps)

    output_eval_file = os.path.join(self.FLAGS.output_dir, "eval_results.txt")
    with tf.gfile.GFile(output_eval_file, "w") as writer:
      tf.logging.info("***** Eval results *****")
      for key in sorted(result.keys()):
        tf.logging.info("  %s = %s", key, str(result[key]))
        writer.write("%s = %s\n" % (key, str(result[key])))



import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import DEP_PARSING_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, DependencyParsingDataFlow
from relogic.logickit.tokenizer.tokenization import BertTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 15), (15, 40), (40, 450)],
    "max_seq_length": 450,
    "label_mapping_path": "data/preprocessed_data/universal_dependency_labels.json"
  })

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased", do_lower_case=False, pretokenized=True),
}
dataflow: DependencyParsingDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[DEP_PARSING_TASK](
  task_name=DEP_PARSING_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=json.load(open(config.label_mapping_path)))

examples = [
{"tokens": ["Al", "-", "Zaman", ":", "American", "forces", "killed", "Shaikh", "Abdullah", "al", "-", "Ani", ",",
            "the", "preacher", "at", "the", "mosque", "in", "the", "town", "of", "Qaim", ",", "near", "the", "Syrian",
            "border", "."],
 "labels": ["root", "punct", "flat", "punct", "amod", "nsubj", "parataxis", "obj", "flat", "flat", "punct", "flat",
            "punct", "det", "appos", "case", "det", "obl", "case", "det", "nmod", "case", "nmod", "punct", "case",
            "det", "amod", "nmod", "punct"],
 "arcs": [0, 1, 1, 1, 6, 7, 1, 7, 8, 8, 8, 8, 8, 15, 8, 18, 18, 7, 21, 21, 18, 23, 21, 21, 28, 28, 28, 21, 1],
 "pos": ["PROPN", "PUNCT", "PROPN", "PUNCT", "ADJ", "NOUN", "VERB", "PROPN", "PROPN", "PROPN", "PUNCT", "PROPN",
         "PUNCT", "DET", "NOUN", "ADP", "DET", "NOUN", "ADP", "DET", "NOUN", "ADP", "PROPN", "PUNCT", "ADP", "DET",
         "ADJ", "NOUN", "PUNCT"]},
{"tokens": ["[", "This", "killing", "of", "a", "respected", "cleric", "will", "be", "causing", "us", "trouble", "for",
            "years", "to", "come", ".", "]"],
 "labels": ["punct", "det", "nsubj", "case", "det", "amod", "nmod", "aux", "aux", "root", "iobj", "obj", "case",
            "obl", "mark", "acl", "punct", "punct"],
 "arcs": [10, 3, 10, 7, 7, 7, 3, 10, 10, 0, 10, 10, 14, 10, 16, 14, 10, 10],
 "pos": ["PUNCT", "DET", "NOUN", "ADP", "DET", "ADJ", "NOUN", "AUX", "AUX", "VERB", "PRON", "NOUN", "ADP", "NOUN",
         "PART", "VERB", "PUNCT", "PUNCT"]}
]

dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=2):
  print(mb)

raise NotImplementedError("You can start to play with data")


from __future__ import absolute_import, division, print_function

import bisect
import time
from relogic.logickit.training.training_scheme import TRAINING_SCHEME
import numpy as np

import torch
from relogic.logickit.base import utils
from relogic.logickit.dataflow import DataFlow
from relogic.logickit.tasks import get_task
from relogic.logickit.model import get_model
from relogic.logickit.training.training_progress import TrainingProgress
from relogic.logickit.dataflow import MiniBatch
import os
from relogic.logickit.utils.utils import print_2d_tensor
from relogic.logickit.base.constants import (
  DISTILL_TASKS, TRAIN_GENERATOR, TRAIN_DISCRIMINATOR, SKIP_EVAL_TASK)
from relogic.logickit.base.configuration import Configuration
from relogic.logickit.tokenizer import NAME_TO_TOKENIZER_MAP

class Trainer(object):
  def __init__(self, config, teacher_config=None,
        tokenizers = None,
        model = None):
    """
    TODO: We are going to change the interface a little bit for easier configuration, less hard-coded.
    :param config:
    :param teacher_config:
    :param tokenizers:
    :param model:
    """
    self.config = config

    ext_config : Configuration = Configuration.load_from_json_file(config.config_file)
    self.ext_config = ext_config

    if tokenizers is not None:
      self.tokenizers = tokenizers
    else:
      self.tokenizers = {}
      for tokenizer_name, params in ext_config.tokenizer_configs.items():
        self.tokenizers[tokenizer_name] = NAME_TO_TOKENIZER_MAP[tokenizer_name].from_pretrained(**params)

    # A quick fix for version migration
    self.tasks = [
      get_task(self.config, task_name, self.tokenizers)
      for task_name in self.config.task_names
    ]
    self.model = get_model(config)(config=self.config, tasks=self.tasks, ext_config=ext_config)

    if self.config.use_external_teacher:
      # Extension for multi-model distillation
      self.teacher_tasks = [
        get_task(teacher_config, task_name, self.tokenizers)
        for task_name in teacher_config.task_names
      ]
      self.teacher_model = get_model(teacher_config)(config=teacher_config, tasks=self.teacher_tasks)

    if config.param_initialization is not None:
      self.initialization_from_pretrained(self.config.param_initialization)


  def initialization_from_pretrained(self, model_path):
    pretrained_dict = torch.load(
      model_path, map_location=lambda storage, location: storage)
    model_dict = self.model.model.state_dict()
    if self.ext_config.param_configs.param_assignment is not None:
      for source_key, target_key in self.ext_config.param_configs.param_assignment.items():
        model_dict[target_key] = pretrained_dict[source_key].to(self.model.device)
        utils.log("Replace {} with {}".format(target_key, source_key))
    self.model.model.load_state_dict(model_dict)
    utils.log("Initialize models Restored from {}".format(model_path))
    utils.log("With keys {}".format(pretrained_dict.keys()))


  def train(self, progress: TrainingProgress):
    if self.config.adversarial_training:
      self.adversarial_train(progress)
    else:
      self._train(progress)

  def adversarial_train(self, progress: TrainingProgress):
    heading = lambda s: utils.heading(s, '(' + self.config.model_name + ')')
    trained_on_sentences = 0
    start_time = time.time()
    generator_supervised_loss_total, generator_supervised_loss_count = 0, 0
    generator_dis_loss_total, generator_dis_loss_count = 0, 0
    discriminator_positive_loss_total, discriminator_positive_loss_count = 0, 0
    discriminator_negative_loss_total, discriminator_negative_loss_count = 0, 0
    real_acc_total, real_acc_count = 0, 0
    fake_acc_total, fake_acc_count = 0, 0
    step = 0

    for turn, labeled_mb, unlabeled_mb in self.get_training_mbs():
      labeled_mb: MiniBatch
      unlabeled_mb: MiniBatch
      if turn == TRAIN_DISCRIMINATOR:
        positive_loss, negative_loss, real_acc, fake_acc = self.model.train_discriminator(labeled_mb, unlabeled_mb)
        discriminator_positive_loss_total += positive_loss
        discriminator_positive_loss_count += 1
        discriminator_negative_loss_total += negative_loss
        discriminator_negative_loss_count += 1
        real_acc_total += real_acc
        real_acc_count += 1
        fake_acc_total += fake_acc
        fake_acc_count += 1
      if turn == TRAIN_GENERATOR:
        supervised_loss, dis_loss = self.model.train_generator(labeled_mb, unlabeled_mb)
        generator_supervised_loss_total += supervised_loss
        generator_supervised_loss_count += 1
        generator_dis_loss_total += dis_loss
        generator_dis_loss_count += 1

      step += 1
      if labeled_mb is not None:
        trained_on_sentences += labeled_mb.size
      if unlabeled_mb is not None:
        trained_on_sentences += unlabeled_mb.size

      # Use simplified version of logging.
      # TODO: Will check if we need to the original version
      if step % self.config.print_every == 0:
        utils.log(
          "step {:} - "
          "generator supervised loss {:.3f} - "
          "generator dis loss {:.3f} - "
          "discriminator positive loss {:.3f} - "
          "discriminator negative loss {:.3f} - "
          "discriminator real accuracy {:.3f} - "
          "discriminator fake accuracy {:.3f} - "
          "{:.1f} sentences per second".format(
            step,
            generator_supervised_loss_total / max(1, generator_supervised_loss_count),
            generator_dis_loss_total / max(1, generator_dis_loss_count),
            discriminator_positive_loss_total / max(1, discriminator_positive_loss_count),
            discriminator_negative_loss_total / max(1, discriminator_negative_loss_count),
            real_acc_total / max(1, real_acc_count),
            fake_acc_total / max(1, fake_acc_count),
            trained_on_sentences / (time.time() - start_time)))
        generator_supervised_loss_total, generator_supervised_loss_count = 0, 0
        generator_dis_loss_total, generator_dis_loss_count = 0, 0
        discriminator_positive_loss_total, discriminator_positive_loss_count = 0, 0
        discriminator_negative_loss_total, discriminator_negative_loss_count = 0, 0
        real_acc_total, real_acc_count = 0, 0
        fake_acc_total, fake_acc_count = 0, 0

      if step % self.config.eval_dev_every == 0:
        heading("EVAL on DEV")
        self.evaluate_all_tasks(progress.history)
        progress.save_if_best_dev_model(self.model.model)
        progress.add_evaluated_step(self.model.global_step_labeled)


      if self.config.early_stop_at > 0 and step >= self.config.early_stop_at:
        utils.log("Early stop at step {}".format(step))
        break




  def _train(self, progress: TrainingProgress):
    heading = lambda s: utils.heading(s, '(' + self.config.model_name + ')')
    trained_on_sentences = 0
    start_time = time.time()
    unsupervised_loss_total, unsupervised_loss_count = 0, 0
    supervised_loss_total, supervised_loss_count = 0, 0
    step = 0
    # self.evaluate_all_tasks(progress.history)
    for mb in self.get_training_mbs():
      if mb.task_name not in DISTILL_TASKS:
        loss = self.model.train_labeled_abstract(mb, step)
        supervised_loss_total += loss
        supervised_loss_count += 1
      else:
        if self.config.use_external_teacher:
          self.teacher_model.run_teacher_abstract(mb)
        else:
          self.model.run_teacher_abstract(mb)
        loss = self.model.train_unlabeled_abstract(mb, step)
        unsupervised_loss_total += loss
        unsupervised_loss_count += 1
        mb.teacher_predictions.clear()

      step += 1
      trained_on_sentences += mb.size


      if self.model.global_step_labeled % self.config.print_every == 0 \
            and not progress.log_in_step(self.model.global_step_labeled):
              # and self.model.global_step_unlabeled % self.config.print_every == 0 \

        # a quick patch here
        # TODO: organize better
        self.model.optimizer.update_loss(supervised_loss_total / max(1, supervised_loss_count))

        utils.log(
          "step supervised {:} - "
          "step unsupervised {:} - "
          "supervised loss: {:.3f} - "
          "unsupervised loss : {:.3f} - "
          "{:.1f} sentences per second".format(
            self.model.global_step_labeled, self.model.global_step_unlabeled,
            supervised_loss_total / max(1, supervised_loss_count),
            unsupervised_loss_total / max(1, unsupervised_loss_count),
            trained_on_sentences / (time.time() - start_time)))
        unsupervised_loss_total, unsupervised_loss_count = 0, 0
        supervised_loss_total, supervised_loss_count = 0, 0
        progress.add_log_step(self.model.global_step_labeled)

      if self.model.global_step_labeled % self.config.eval_dev_every == 0 \
            and not progress.evaluated_in_step(self.model.global_step_labeled):
            # and self.model.global_step_unlabeled % self.config.eval_dev_every == 0 and \

        heading("EVAL on DEV")
        self.evaluate_all_tasks(progress.history)
        progress.save_if_best_dev_model(self.model.model)
        progress.add_evaluated_step(self.model.global_step_labeled)

      if self.config.early_stop_at > 0 and self.model.global_step_labeled >= self.config.early_stop_at:
        utils.log("Early stop at step {}".format(self.model.global_step_labeled))
        break

  def evaluate_all_tasks(self, history=None, train_set=False):
    results = []
    for task in self.tasks:
      if task.name in SKIP_EVAL_TASK:
        continue
      results.append(self.evaluate_task(task, train_set))
      if history is not None:
        results[-1].append(('step', self.model.global_step_labeled))
    if history is not None:
      history.append(results)
    if history is not None:
      utils.write_pickle(history, self.config.history_file)

  def evaluate_task(self, task, train_set):
    scorer = task.get_scorer(dump_to_file={"output_dir": self.config.output_dir,
                                           "task_name": task.name})
    data: DataFlow = task.train_set if train_set else task.val_set
    for i, mb in enumerate(data.get_minibatches(self.config.tasks[data.task_name]["test_batch_size"])):
      # batch_preds = self.model.test(mb)
      outputs = self.model.test_abstract(mb)
      task_outputs = outputs[task.name]
      scorer.update(mb, task_outputs, task_outputs.get("loss", 0), {})
      # TODO: This code will break a lot!
      # TODO: But we need to migrate this!

      # if isinstance(batch_preds, dict):
      #   scorer.update(mb, batch_preds, 0, None)
      # else:
      #   # Slow Migration towards the return interface
      #   extra_output = {}
      #   if self.config.output_attentions:
      #     batch_preds, attention_map = batch_preds
      #     extra_output["attention_map"] = attention_map
      #   if isinstance(batch_preds, tuple):
      #     loss, batch_preds = batch_preds
      #   else:
      #     loss = 0
      #   scorer.update(mb, batch_preds, loss, extra_output)
      if i % 100 == 0:
        utils.log("{} batch processed.".format(i))
    results = scorer.get_results()
    utils.log(task.name.upper() + ": " + scorer.results_str())
    return results

  def analyze_task(self, task, head_mask,
                   head_importance, attn_entropy):
    scorer = task.get_scorer(dump_to_file={"output_dir": self.config.output_dir,
                                           "task_name": task.name})
    params = {
      "head_importance": head_importance,
      "attn_entropy": attn_entropy,
      "total_token": 0.0
    }
    data : DataFlow = task.val_set
    # The output of the model is logits and attention_map
    for i, mb in enumerate(data.get_minibatches(self.config.tasks[data.task_name]["test_batch_size"])):
      batch_preds, attention_map = self.model.analyze(mb, head_mask, params=params)
      extra_output = {}
      extra_output["attention_map"] = attention_map
      loss = 0
      scorer.update(mb, batch_preds, loss, extra_output)
      if i % 100 == 0:
        utils.log("{} batch processed.".format(i))
    results = scorer.get_results()
    utils.log(task.name.upper() + ": " + scorer.results_str())

    params["attn_entropy"] /= params["total_token"]
    params["head_importance"] /= params["total_token"]
    np.save(os.path.join(self.config.output_dir, 'attn_entropy.npy'), attn_entropy.detach().cpu().numpy())
    np.save(os.path.join(self.config.output_dir, 'head_importance.npy'), head_importance.detach().cpu().numpy())

    utils.log("Attention entropies")
    print_2d_tensor(attn_entropy)
    utils.log("Head importance scores")
    print_2d_tensor(head_importance)
    utils.log("Head ranked by importance scores")
    head_ranks = torch.zeros(head_importance.numel(), dtype=torch.long, device=self.model.device)
    head_ranks[head_importance.view(-1).sort(descending=True)[1]] = torch.arange(head_importance.numel(),
                                                                                 device=self.model.device)
    head_ranks = head_ranks.view_as(head_importance)
    print_2d_tensor(head_ranks)
    return results

  def get_training_mbs(self):
    # training_scheme_config = self.ext_config.training_scheme
    if self.config.training_scheme is not None:
      yield from TRAINING_SCHEME[self.config.training_scheme](self.config, self.tasks)
    else:
      datasets = [task.train_set for task in self.tasks]
      weights = [np.sqrt(dataset.size) for dataset in datasets]
      thresholds = np.cumsum([w / np.sum(weights) for w in weights])

      labeled_mbs = [
        dataset.endless_minibatches(self.config.tasks[dataset.task_name]["train_batch_size"], bucket=not self.config.no_bucket)
        for dataset in datasets
      ]
      while True:
        dataset_ind = bisect.bisect(thresholds, np.random.random())
        yield next(labeled_mbs[dataset_ind])


  def restore(self, model_path):
    restore_state_dict = torch.load(
      model_path, map_location=lambda storage, location: storage)
    # loaded_dict = {k: restore_state_dict[k] for k in
    #                set(self.model.model.state_dict().keys()) & set(restore_state_dict.keys())}
    # model_state = self.model.model.state_dict()
    # model_state.update(loaded_dict)
    for key in self.config.ignore_parameters:
      # restore_state_dict.pop(key)
      restore_state_dict[key] = self.model.model.state_dict()[key]
    self.model.model.load_state_dict(restore_state_dict)
    utils.log("Model Restored from {}".format(model_path))

  def restore_teacher(self, model_path):
    restore_state_dict = torch.load(
      model_path, map_location=lambda storage, location: storage)
    # loaded_dict = {k: restore_state_dict[k] for k in
    #                set(self.model.model.state_dict().keys()) & set(restore_state_dict.keys())}
    # model_state = self.model.model.state_dict()
    # model_state.update(loaded_dict)
    for key in self.config.ignore_parameters:
      # restore_state_dict.pop(key)
      restore_state_dict[key] = self.model.model.state_dict()[key]
    self.teacher_model.model.load_state_dict(restore_state_dict)
    utils.log("Teacher Model Restored from {}".format(model_path))

  def feature_extraction(self, task, dump_file):
    data = task.val_set
    fout = open(dump_file, 'wb')
    features = []
    for i, mb in enumerate(data.get_minibatches(self.config.tasks[data.task_name]["test_batch_size"])):
      # batch_preds = self.model.test(mb)
      batch_preds = self.model.test_abstract(mb)
      features.append(batch_preds[task.name]["features"].cpu().detach())
      if i % 100 == 0:
        utils.log("{} batch processed.".format(i))
    features = torch.cat(features, dim=0).numpy()
    np.savez(fout, X=np.array(features))
    fout.close()


from flask import Flask, jsonify, request


app = Flask(__name__)

@app.route('/')
def index():
  return jsonify("Hello World")

class Server(object):
  def __init__(self, trainer=None):
    self.trainer = trainer

  def start(self):
    app.run(host="0.0.0.0", port=5000, debug=True)


if __name__ == "__main__":
  server = Server()
  server.start()


from typing import List

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.tokenization import BertTokenizer
from relogic.logickit.utils import create_tensor, truncate_seq_pair


class DocPointwiseExample(Example):
  """

  """
  def __init__(self, guid: str, text_a: str, text_bs: List[str], label=None, sent_label=None):
    super().__init__()
    self.guid = guid
    self.text_a = text_a
    self.text_bs = text_bs
    self.label = label
    self.sent_label = sent_label
    self.sent_num = len(self.text_bs)

  def process(self, tokenizers, *inputs, regression=False, **kwargs):
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, BertTokenizer):
        self.text_a_tokens, self.text_a_is_head = tokenizer.tokenize(self.text_a)
        self.text_bs_tokens, self.text_bs_is_head = zip(*[tokenizer.tokenize(text_b) for text_b in self.text_bs])

        max_seq_length = kwargs.pop("max_seq_length", 512)
        for i in range(self.sent_num):
          truncate_seq_pair(self.text_a_tokens, self.text_bs_tokens[i], max_seq_length - 3)
          truncate_seq_pair(self.text_a_is_head, self.text_bs_is_head[i], max_seq_length - 3)
        self.tokens_list = []
        self.segment_ids_list = []
        self.is_head_list = []
        self.input_ids_list = []
        self.input_mask_list = []
        for i in range(self.sent_num):
          self.tokens_list.append(["[CLS]"] + self.text_a_tokens + ["[SEP]"] + self.text_bs_tokens[i] + ["[SEP]"])
          self.segment_ids_list.append([0] * (len(self.text_a_tokens) + 2) + [1] * (len(self.text_bs_tokens[i]) + 1))
          self.is_head_list.append([2] + self.text_a_is_head + [2] + self.text_bs_is_head[i] + [2])

          self.input_ids_list.append(tokenizer.convert_tokens_to_ids(self.tokens_list[i]))
          self.input_mask_list.append([1] * len(self.input_ids_list[i]))

        if self.label is not None:
          if regression:
            # Regression Problem with sigmoid loss function
            self.label_ids = float(self.label)
          else:
            label_mapping = kwargs.get("label_mapping")
            self.label_ids = label_mapping[self.label]

  @classmethod
  def from_structure(cls, structure):
    return cls(guid="", text_a=structure.text_a, text_bs=structure.text_b)

  @classmethod
  def from_json(cls, example):
    return cls(guid="{}|{}".format(example.get("text_a_id", 0), example.get("text_b_id", 0)),
               text_a=example["text_a"],
               text_bs=example["text_b"],
               label=example.get("label", None))

  @property
  def len(self):
    return 1

  def _len(self, i):
    return len(self.input_ids_list[i])

class DocPointwiseFeature(Feature):
  def __init__(self, *inputs, **kwargs):
    super().__init__()
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.label_ids = kwargs.pop("label_ids")

class DocPointwiseMiniBatch(MiniBatch):
  def __init__(self, *inputs, **kwargs):
    super().__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                         torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["input_head"] = create_tensor(self.input_features, "is_head",
                                         torch.long, device)
    if use_label:
      if self.config.regression:
        inputs["label_ids"] = create_tensor(self.input_features, "label_ids",
                                            torch.float, device)
      else:
        inputs["label_ids"] = create_tensor(self.input_features, "label_ids",
                                            torch.long, device)
    else:
      inputs["label_ids"] = None
    inputs["extra_args"] = {}
    return inputs

class DocPointwiseDataFlow(DataFlow):
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super().__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return DocPointwiseExample

  @property
  def minibatch_class(self):
    return DocPointwiseMiniBatch

  def process_example(self, example: DocPointwiseExample):
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping,
                    max_seq_length=self.config.max_seq_length,
                    regression=self.config.regression)

  def convert_examples_to_features(self, examples: List[DocPointwiseExample]):
    examples: List[DocPointwiseExample]
    # Because we want to have batch operation on one example
    # So len(examples) == 1
    features = []
    assert len(examples) == 1, "We can only handle one example at a time for batching"
    example = examples[0]
    max_token_length = max([example._len(i) for i in range(example.sent_num)])

    for idx in range(example.sent_num):
      padding = [0] * (max_token_length - example._len(idx))
      input_ids = example.input_ids_list[idx] + padding
      input_mask = example.input_mask_list[idx] + padding
      segment_ids = example.segment_ids_list[idx] + padding
      if hasattr(example, "label_ids"):
        label_ids = example.label_ids
      else:
        label_ids = None

      features.append(
        DocPointwiseFeature(input_ids=input_ids,
                         input_mask=input_mask,
                         segment_ids=segment_ids,
                         label_ids=label_ids))
    return features

  def decode_to_labels(self, preds, mbs: DocPointwiseMiniBatch):
    return None


from dataclasses import dataclass, field
from typing import List

from relogic.structures.structure import Structure
from relogic.structures.paragraph import Paragraph
from relogic.structures.sentence import Sentence


@dataclass
class Document(Structure):
  idx: int = None
  text: str = None
  paragraphs: List[Paragraph] = field(default_factory=list)

  tokenizer: str = "space"

  def __post_init__(self):
    for para in self.text.split("\n"):
      self.add_paragraph(Paragraph(text=para, tokenizer=self.tokenizer))

  def add_paragraph(self, paragraph: Paragraph):
    self.paragraphs.append(paragraph)




import torch.nn as nn
import torch.nn.functional as F
import torch
from relogic.logickit.base import utils
from relogic.logickit.utils import utils

class SelectIndexModule(nn.Module):
  def __init__(self, config, task_name, n_classes=None):
    super(SelectIndexModule, self).__init__()
    self.config = config
    self.task_name = task_name

  def forward(self, *inputs, **kwargs):

    student_results = kwargs.pop("student_results")
    teacher_results = kwargs.pop("teacher_results")
    # a_selected_indices = kwargs.pop("a_selected_indices")
    # b_selected_indices = kwargs.pop("b_selected_indices")

    if self.config.select_index_method == "avg":
      # # a is the teacher
      # teacher_mask = (kwargs["a_input_ids"] > 0).float()
      # student_mask = (kwargs["input_ids"] > 0).float()
      # # 101 CLS 102 SEP
      # teacher_CLS_mask = (kwargs["a_input_ids"] == 101).float()
      # teacher_SEP_mask = (kwargs["a_input_ids"] == 102).float()
      # student_CLS_mask = (kwargs["input_ids"] == 101).float()
      # student_SEP_mask = (kwargs["input_ids"] == 102).float()
      # teacher_length = torch.sum(kwargs["a_input_ids"] > 0, dim=-1) - 2
      # student_length = torch.sum(kwargs["input_ids"] > 0, dim=-1) - 2
      # # Remove the CLS and SEP

      # teacher_sum = torch.sum(teacher_results * (teacher_mask - teacher_CLS_mask - teacher_SEP_mask).unsqueeze(-1), -2)
      # student_sum = torch.sum(student_results * (student_mask - student_CLS_mask - student_SEP_mask).unsqueeze(-1), -2)
      # teacher_avg = teacher_sum / teacher_length.unsqueeze(-1).float()
      # student_avg = student_sum / student_length.unsqueeze(-1).float()

      teacher_mask = (kwargs["a_is_head"] == 1).float()
      student_mask = (kwargs["b_is_head"] == 1).float()
      teacher_length = torch.sum(teacher_mask, dim=-1)
      student_length = torch.sum(student_mask, dim=-1)
      teacher_sum = torch.sum(teacher_results * teacher_mask.unsqueeze(-1), -2)
      student_sum = torch.sum(student_results * student_mask.unsqueeze(-1), -2)
      teacher_avg = teacher_sum / teacher_length.unsqueeze(-1).float()
      student_avg = student_sum / student_length.unsqueeze(-1).float()



      return student_avg, teacher_avg, None
    elif self.config.select_index_method == "cls":
      a_selected_features = teacher_results[:, 0]
      b_selected_features = student_results[:, 0]
      return b_selected_features, a_selected_features, None
    else:
      raise NotImplementedError("The selection method is not defined")
    # selected_indices_mask = (a_selected_indices > 0)

    # a_selected_features = utils.batched_index_select(teacher_results, a_selected_indices)
    # b_selected_features = utils.batched_index_select(student_results, b_selected_indices)

    # teacher_seq_length = teacher_results.size(1)
    # student_seq_length = student_results.size(1)
    # # batch_size, sequence_length, dim -> batch_size, dim, sequence_length
    # a_maxpooling_features = F.max_pool1d(teacher_results[:, 1:].transpose(1, 2), kernel_size=teacher_seq_length-1).squeeze(-1)
    # b_maxpooling_features = F.max_pool1d(student_results[:, 1:].transpose(1, 2), kernel_size=student_seq_length-1).squeeze(-1)



    # return b_selected_features, a_selected_features, selected_indices_mask

    # return b_maxpooling_features, a_maxpooling_features, None


from relogic.logickit.tokenizer.tokenization import BertTokenizer
from relogic.logickit.tasks import get_task
from relogic.logickit.tasks.task import Task
from relogic.logickit.model import get_model
from relogic.logickit.dataflow import DataFlow
import torch
from relogic.logickit.base import utils
from relogic.logickit.base.configuration import Configuration
from relogic.logickit.tokenizer import NAME_TO_TOKENIZER_MAP
import os

class Predictor(object):
  def __init__(self, config, tokenizers=None):
    self.config = config
    ext_config: Configuration = Configuration.load_from_json_file(config.config_file)
    self.ext_config = ext_config

    if tokenizers is not None:
      self.tokenizers = tokenizers
    else:
      self.tokenizers = {}
      for tokenizer_name, params in ext_config.tokenizer_configs.items():
        self.tokenizers[tokenizer_name] = NAME_TO_TOKENIZER_MAP[tokenizer_name].from_pretrained(**params)

    self.tasks = [
      get_task(self.config, task_name, self.tokenizers)
      for task_name in self.config.task_names
    ]
    # self.tokenizer = BertTokenizer.from_pretrained(
    #   config.vocab_path, do_lower_case=config.do_lower_case,
    #   never_split=config.never_split, lang=config.lang)
    # self.tasks = [
    #   get_task(self.config, task_name, self.tokenizer)
    #   for task_name in self.config.task_names
    # ]
    self.model = get_model(config)(config=self.config, tasks=self.tasks, ext_config=ext_config)

  def predict(self, structures, selected_task=None):
    # We currently assume it is single task model
    sel_task = None
    if selected_task is not None:
      for task in self.tasks:
        if task.name == selected_task:
          sel_task = task
    task: Task = self.tasks[0] if selected_task is None else sel_task
    data: DataFlow = task.dataset
    data.update_with_structures(structures)
    # results = []
    for i, mb in enumerate(data.get_minibatches(self.config.tasks[task.name]["test_batch_size"])):
      # batch_preds = self.model.test(mb)
      batch_preds = self.model.test_abstract(mb)
      # batch_preds = batch_preds[task.name]["logits"]
      labels = data.decode_to_labels(batch_preds, mb)
      yield batch_preds, labels, mb
      # results.append(batch_preds.data.cpu().numpy())
      # labels = data.decode_to_labels(batch_preds)
    #   if i % 100 == 0:
    #     print("Processed {} batches".format(i))
    # return results

  def restore(self, model_path, model_name="default.ckpt"):
    model_ckpt_path = os.path.join(model_path, model_name)
    restore_state_dict = torch.load(
      model_ckpt_path, map_location=lambda storage, location: storage)
    # loaded_dict = {k: restore_state_dict[k] for k in
    #                set(self.model.model.state_dict().keys()) & set(restore_state_dict.keys())}
    # model_state = self.model.model.state_dict()
    # model_state.update(loaded_dict)
    for key in self.config.ignore_parameters:
      # restore_state_dict.pop(key)
      restore_state_dict[key] = self.model.model.state_dict()[key]
    self.model.model.load_state_dict(restore_state_dict)
    utils.log("Model Restored from {}".format(model_path))






import abc
import torch

class Scorer(object, metaclass=abc.ABCMeta):
  def __init__(self):
    self._updated = False
    self._cached_results = {}
    self.results = None
    self.need_to_clear_output = False
    self.dump_to_file_path = None
    self.dump_to_file_handler = None


  @abc.abstractmethod
  def update(self, mbs, predictions, loss, extra_args):
    self._updated = True


  @abc.abstractmethod
  def get_loss(self):
    pass

  @abc.abstractmethod
  def _get_results(self):
    return []

  def get_results(self):
    if self.dump_to_file_path:
      self.need_to_clear_output = True
    results = self._get_results()
    self.results = results
    return results

  def results_str(self):
    return " - ".join(["{:}: {:.2f}".format(k, v)
                       for k, v in self.results])

  @staticmethod
  def unwrap_to_tensors(*tensors: torch.Tensor):
    """
    If you actually passed gradient-tracking Tensors to a Metric, there will be
    a huge memory leak, because it will prevent garbage collection for the computation
    graph. This method ensures that you're using tensors directly and that they are on
    the CPU.
    """
    return (x.detach().cpu() if isinstance(x, torch.Tensor) else x for x in tensors)


from relogic.logickit.tasks.task import Task
from relogic.logickit.modules.srl_module import SRLModule, PipeSRLModule
from relogic.logickit.modules.tagging_module import TaggingModule
from relogic.logickit.modules.predicate_sense_module import PredicateSenseModule
from relogic.logickit.modules.joint_srl_module import JointSRLModule
from relogic.logickit.modules.sequence_labeling_module import SequenceLabelingModule
from relogic.logickit.scorer.tagging_scorers import EntityLevelF1Scorer, AccuracyScorer
from relogic.logickit.scorer.srl_scorers import SRLF1Scorer, SpanSRLF1Scorer, JointSpanSRLF1Scorer
from relogic.logickit.dataset.labeled_data_loader import LabeledDataLoader
from relogic.logickit.base.constants import (SEQUENCE_LABELING_TASK, NER_TASK, POS_TASK, PIPE_SRL_TASK,
                                             PREDICATE_DETECTION_TASK)

class Tagging(Task):
  def __init__(self, config, name,  tokenizer=None):
    super(Tagging, self).__init__(
      config, name, LabeledDataLoader(config, name, tokenizer))
    self.n_classes = len(set(self.loader.label_mapping.values()))

  def get_module(self):
    if self.name in ['srl', "srl_conll05", "srl_conll09", "srl_conll12"]:
      # if not self.config.predicate_surface_aware:
      #   return
      return SRLModule(self.config, self.name, self.n_classes)
    if self.name in ["joint_srl"]:
      return JointSRLModule(self.config, self.name, self.n_classes)
    elif self.name in ['predicate_sense']:
      return PredicateSenseModule(self.config, self.name, self.n_classes)
    elif self.name in [NER_TASK, POS_TASK, PREDICATE_DETECTION_TASK]:
      return SequenceLabelingModule(self.config, self.name, self.n_classes)
    elif self.name in [PIPE_SRL_TASK]:
      return PipeSRLModule(self.config, self.name, self.n_classes)
    else:
      return TaggingModule(self.config, self.name, self.n_classes)
      # raise ValueError("Task name {} is not defined".format(self.name))

  def get_scorer(self, dump_to_file=None):
    if self.name in ["er", "ner", NER_TASK, PREDICATE_DETECTION_TASK]:
      return EntityLevelF1Scorer(label_mapping=self.loader.label_mapping, dump_to_file=dump_to_file)
    if self.name in ["srl", "srl_conll05", "srl_conll09", "srl_conll12", PIPE_SRL_TASK]:
      if self.config.span_inference:
        return SpanSRLF1Scorer(label_mapping=self.loader.label_mapping, dump_to_file=dump_to_file)
      else:
        return SRLF1Scorer(label_mapping=self.loader.label_mapping, dump_to_file=dump_to_file)
    if self.name in ["joint_srl"]:
      return JointSpanSRLF1Scorer(label_mapping=self.loader.label_mapping, dump_to_file=dump_to_file)
    if self.name in ["predicate_sense"]:
      return AccuracyScorer(ignore_list=['O'], label_mapping=self.loader.label_mapping, dump_to_file=dump_to_file)
    if self.name in [POS_TASK]:
      return AccuracyScorer(ignore_list=['_'], label_mapping=self.loader.label_mapping, dump_to_file=dump_to_file)


from typing import List
from types import SimpleNamespace

import torch

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import PAIRWISE_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, PairwiseDataFlow
from relogic.logickit.tokenizer.tokenization import BertTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 15), (15, 40), (40, 450)],
    "max_seq_length": 450
  }
)

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased", do_lower_case=True, lang="zh"),
}

dataflow: PairwiseDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[PAIRWISE_TASK](
  task_name=PAIRWISE_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=None)

examples = [{
  "guid": "1",
  "text": "林郑月娥，香港特別行政区現任行政長官，在36年間曾於20個政府不同的工作崗位上工作。",
  "text_p": "Carrie Lam Cheng Yuet-ngor is a Hong Kong politician serving as the 4th Chief Executive of Hong Kong since 2017.",
  "text_n": "Leung Chun-ying, also known as CY Leung, is a Hong Kong politician. He served as the third Chief Executive of Hong Kong between 2012 and 2017."
}, {
  "guid": "2",
  "text": "天安门广场是位于北京市中心的城市广场，是世界上最大的城市广场之一。因位于明清北京皇城的南门「天安门」外而得名。广场北端设有国旗杆，每天都會隨日出、日落進行升旗、降旗仪式。",
  "text_p": "Tiananmen Square or Tian'anmen Square is a city square in the centre of Beijing, China, named after the Tiananmen located to its north, separating it from the Forbidden City.",
  "text_n": "The Forbidden City is a palace complex in central Beijing, China. It houses the Palace Museum, and was the former Chinese imperial palace from the Ming dynasty to the end of the Qing dynasty (the years 1420 to 1912)."
}]

dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=2):
  print(mb)

raise NotImplementedError("You can start to play with data")


"""
This module contains the basic components of dataflow,
including Example, Feature, MiniBatch and DataFlow classes.

"""

import abc
import collections
import json
import random
from typing import Dict

import numpy as np

from relogic.logickit.base import utils


class Example(object, metaclass=abc.ABCMeta):
  """Basic Example class."""

  def __init__(self):
    pass

  @abc.abstractmethod
  def process(self, tokenizer, *inputs, **kwargs):
    """Process function abstract. Need to be implemented in each Subclass"""
    raise NotImplementedError()

  @property
  def len(self):
    raise NotImplementedError()

  @property
  def _len(self):
    raise NotImplementedError()

  @property
  def bucketing_len(self):
    try:
      return self.len
    except:
      return self._len


class Feature(object, metaclass=abc.ABCMeta):
  """Basic Feature class."""

  def __init__(self):
    pass


class MiniBatch(object, metaclass=abc.ABCMeta):
  """Basic MiniBatch class.

  For minimal requirements, this class needs to have the following arguments.

  Args:
    config : The configuration of the DataFlow to which it belongs.
    examples (List[Example]): A list of examples.
    size (int): The size of the batch.
    input_features (List[Feature]): A list of features (each of them are padded).
    teacher_predictions (Dict): This is for semi-supervised learning.

  """

  def __init__(self, *inputs, **kwargs):
    self.task_name = kwargs.pop("task_name")
    self.config = kwargs.pop("config")
    self.examples = kwargs.pop("examples")
    self.size = kwargs.pop("size")
    self.input_features = kwargs.pop("input_features")
    self.teacher_predictions = kwargs.pop("teacher_predictions")
    if hasattr(self.config, "tasks") and "loss_weight" in self.config.tasks[self.task_name]:
      self.loss_weight = self.config.tasks[self.task_name]["loss_weight"]
    else:
      self.loss_weight = 1

  @abc.abstractmethod
  def generate_input(self, device, use_label):
    """Convert the features to tensors

    Args:
      device (torch.Device): gpu device or cpu.
      use_label (bool): To create the label tensor or not.

    """
    raise NotImplementedError()


def get_bucket(config, length):
  """classify each example into given buckets with its length."""

  for i, (start, end) in enumerate(config.buckets):
    if start <= length < end:
      return config.buckets[i]

  return config.buckets[-1]


class DataFlow(object, metaclass=abc.ABCMeta):
  """DataFlow controls the data process and batch generation.

  The DataFlow adopts examples from Structure or json object.

  Note: Most current implementation is based on BERT model.

  Args:
    config (SimpleNamespace): Configuration for the DataFlow class.
    tokenizer: Tokenizer for string tokenization.

  """

  def __init__(self, config, task_name, tokenizers: Dict, label_mapping):
    self.config = config
    self.task_name = task_name
    self.tokenizers = tokenizers
    self.examples = []
    self.label_mapping = label_mapping

    # if label_mapping_path == "none":
    #   self.label_mapping = {}
    # else:
    #   self.label_mapping = json.load(open(label_mapping_path))

  @abc.abstractmethod
  def process_example(self, example):
    """Basic method for example processing. This method needs be implemented
    case by case. For different Subclass, it has different arguments during
    the example processing.

    """
    raise NotImplementedError()

  def update_with_structures(self, structures):
    """Convert the Structure into Example.

    This method is used during the deployment.

    Args:
      structures (List[Structure]): List of Structure.

    """
    self.examples = [
        self.example_class.from_structure(structure)
        for structure in structures
    ]
    from tqdm import tqdm
    for idx, example in tqdm(enumerate(self.examples), desc="Loading data", total=len(self.examples)):
      self.process_example(example)

  def update_with_jsons(self, examples):
    """Convert json object into Example.

    This method can be used in deployment or training.

    Args:
      examples: (List[Dict]): List of json objects.

    """
    self.examples = [
        self.example_class.from_json(example) for example in examples
    ]
    from tqdm import tqdm
    for idx, example in tqdm(enumerate(self.examples), desc="Loading data", total=len(self.examples)):
      self.process_example(example)
      if idx < 3:
        print(example.__dict__)

  def update_with_file(self, file_name):
    """Read json objects from file.

    Args:
      file_name (str): Filename.

    """
    examples = []
    with open(file_name) as fin:
      for line in fin:
        examples.append(json.loads(line))
    self.update_with_jsons(examples)

  def endless_minibatches(self, minibatch_size, sequential=False, bucket=True):
    """Generate endless minibatches with given batch size."""

    if sequential:
      print("Use sequential dataset for {}".format(self.task_name))
    elif not bucket:
      print("Use random shuffle dataset for {}".format(self.task_name))
    else:
      print("Use bucket dataset for {}".format(self.task_name))
    while True:
      for minibatch in self.get_minibatches(minibatch_size, sequential=sequential, bucket=bucket):
        yield minibatch

  def get_minibatches(self, minibatch_size, sequential=True, bucket=False):
    """Generate list of batch size based on examples.

    There are two modes for generating batches. One is sequential,
    which follows the original example sequence in the dataset.
    The other mode is based on bucketing, to save the memory consumption.

    NOTICE: The default value here is for evaluation. Because the evaluation interface will directly call this.
    The traning interface will call endless_minibatches.

    Args:
      minibatch_size (int): Batch size.
      sequential (bool): To be sequential mode or not.

    """
    if sequential:
      index = 0
      while index < self.size:
        yield self._make_minibatch(
            np.array(range(index, min(index + minibatch_size, self.size))))
        index += minibatch_size
    elif not bucket:
      indices = list(range(self.size))
      random.shuffle(indices)
      indices = np.array(indices)
      index = 0
      while index < self.size:
        yield self._make_minibatch(
            indices[index: min(index + minibatch_size, self.size)])
        index += minibatch_size
    else:
      by_bucket = collections.defaultdict(list)
      for i, example in enumerate(self.examples):
        by_bucket[get_bucket(self.config, example.bucketing_len)].append(i)
      # save memory by weighting examples so longer sentences
      #   have smaller minibatches
      weight = lambda ind: np.sqrt(self.examples[ind].bucketing_len)
      total_weight = float(sum(weight(i) for i in range(self.size)))
      weight_per_batch = minibatch_size * total_weight / self.size
      cumulative_weight = 0.0
      id_batches = []
      for _, ids in by_bucket.items():
        ids = np.array(ids)
        np.random.shuffle(ids)
        curr_batch, curr_weight = [], 0.0
        for i, curr_id in enumerate(ids):
          curr_batch.append(curr_id)
          curr_weight += weight(curr_id)
          if (i == len(ids) - 1 or cumulative_weight + curr_weight >=
              (len(id_batches) + 1) * weight_per_batch):
            cumulative_weight += curr_weight
            id_batches.append(np.array(curr_batch))
            curr_batch, curr_weight = [], 0.0
      random.shuffle(id_batches)
      utils.log("Data Flow {}, There are {} batches".format(
          self.__class__, len(id_batches)))
      for id_batch in id_batches:
        yield self._make_minibatch(id_batch)

  def _make_minibatch(self, ids):
    """Make a MiniBatch given ids.

    Given ids, the method chooses the corresponding examples.

    Args:
      ids (List(int)): id list.

    Return:
      MiniBatch: The created Minibatch.
    """
    examples = [self.examples[i] for i in ids]
    input_features = self.convert_examples_to_features(examples=examples)

    return self.minibatch_class(
        task_name=self.task_name,
        config=self.config,
        size=ids.size,
        examples=examples,
        teacher_predictions={},
        input_features=input_features)

  @abc.abstractmethod
  def convert_examples_to_features(self, examples):
    """Basic method abstraction for converting examples to features."""

    raise NotImplementedError()

  @property
  @abc.abstractmethod
  def example_class(self):
    """Return the Example class based on the Subclass."""

    raise NotImplementedError()

  @property
  def minibatch_class(self):
    """Return the MiniBatch class based on the Subclass."""

    raise NotImplementedError()

  @property
  def size(self):
    """The size of the dataset."""

    return len(self.examples)



import torch.nn as nn
import copy
import torch
from relogic.logickit.modules.rnn import dynamic_rnn


# class PredictionModule(nn.Module):
#   def __init__(self, config, task_name, n_classes, activate=True):
#     super(PredictionModule, self).__init__()
#     self.config = config
#     self.task_name = task_name
#     self.n_classes = n_classes
#     # self.projection = nn.Linear(config.hidden_size, config.projection_size)
#     # self.activate = activate
#     # if activate:
#     #   self.activation = nn.ReLU()
#     self.to_logits = nn.Linear(config.hidden_size, self.n_classes)
#     self.apply(self.init_weights)
#
#   def init_weights(self, module):
#     if isinstance(module, nn.Linear):
#       module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
#     if isinstance(module, nn.Linear) and module.bias is not None:
#       module.bias.data.zero_()
#
#   def forward(self, input):
#     # projected = self.projection(input)
#     # if self.activate:
#     #   projected = self.activation(projected)
#     logits = self.to_logits(input)
#     return logits

class PredictionModule(nn.Module):
  def __init__(self, config, task_name, n_classes, activate=True):
    super(PredictionModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes
    if self.config.use_bilstm:
      self.projection = nn.Linear(2 * config.hidden_size, config.hidden_size)
      self.activate = activate
      if activate:
        self.activation = nn.SELU()
    self.to_logits = nn.Linear(config.hidden_size, self.n_classes)
    # self.apply(self.init_weights)

  def init_weights(self, module):
    if isinstance(module, nn.Linear):
      module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
    if isinstance(module, nn.Linear) and module.bias is not None:
      module.bias.data.zero_()

  def forward(self, input):
    if self.config.use_bilstm:
      projected = self.projection(input)
      if self.activate:
        projected = self.activation(projected)
    else:
      projected = input
    logits = self.to_logits(projected)
    return logits

class PartialViewPredictionModule(nn.Module):
  def __init__(self, config, task_name, n_classes, activate=True):
    super(PartialViewPredictionModule, self).__init__()
    self.config = config

    self.projection = nn.Linear(config.hidden_size, config.hidden_size)
    self.activate = activate
    if activate:
      self.activation = nn.SELU()
    self.to_logits = nn.Linear(config.hidden_size, n_classes)
    # self.apply(self.init_weights)

  def init_weights(self, module):
    if isinstance(module, nn.Linear):
      module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
    if isinstance(module, nn.Linear) and module.bias is not None:
      module.bias.data.zero_()

  def forward(self, input):
    projected = self.projection(input)
    if self.activate:
      projected = self.activation(projected)
    logits = self.to_logits(projected)
    return logits

class TaggingModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(TaggingModule, self).__init__()
    self.config = config
    self.partial_view_sources = config.partial_view_sources
    if self.config.use_bilstm:
      self.bilstm = nn.LSTM(
        input_size=config.hidden_size,  # for indicator embedding
        hidden_size=config.hidden_size,
        bidirectional=True,
        batch_first=True,
        num_layers=1)
    self.primary = PredictionModule(config, task_name, n_classes)
    # self.partial_list = nn.ModuleList([
    #   PredictionModule(config, task_name, n_classes)
    #   for _ in range(len(config.partial_view_sources))])
    if self.config.is_semisup:
      self.partial_list = nn.ModuleList([
        PartialViewPredictionModule(config, task_name, n_classes)
        for _ in range(4)])

  def padding(self, batch, dim, device):
    return torch.zeros([batch, 1, dim], device=device).float()

  def forward(self,
              input,
              view_type="primary",
              final_multi_head_repr=None,
              input_mask=None,
              segment_ids=None,
              extra_args=None,
              **kwargs):
    # print(input_mask)
    input_lengths = (input_mask == 1).sum(-1)
    if self.config.use_bilstm:
      output, (ht, ct) = dynamic_rnn(self.bilstm, input, input_lengths)
    else:
      output = input
    if view_type == "primary":
      primary_logits = self.primary(output)
      return primary_logits
    else:
      # if final_multi_head_repr is None:
      #   raise ValueError("final_multi_head_repr should have value instead of None")
      # new_shape_final_multi_head_repr = final_multi_head_repr.size()[:-1] + \
      #    (int(final_multi_head_repr.size(-1) / self.config.self_attention_head_size), self.config.self_attention_head_size, )
      # final_multi_head_repr = final_multi_head_repr.view(*new_shape_final_multi_head_repr)
      partial_view_results = []
      padding = self.padding(output.size(0), self.config.hidden_size, output.device)
      for idx, partial_view_module in enumerate(self.partial_list):
        # partial_view_results.append(partial_view_module(input[source_layer]))
        # partial_view_results.append(partial_view_module(final_multi_head_repr[:, :, head_index, :]))
        if idx == 0:
          partial_view_results.append(partial_view_module(output[:,:,:self.config.hidden_size]))
        if idx == 1:
          partial_view_results.append(partial_view_module(output[:,:,self.config.hidden_size:]))
        if idx == 2:
          partial_view_results.append(
            partial_view_module(torch.cat([padding, output[:,:-1,:self.config.hidden_size]], dim=1)))
        if idx == 3:
          partial_view_results.append(
            partial_view_module(torch.cat([output[:,1:,self.config.hidden_size:], padding], dim=1)))
      return partial_view_results


import torch.nn as nn
import torch
class PredicateSenseModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(PredicateSenseModule, self).__init__()
    self.config = config
    self.predicate_indicator_embedding = nn.Embedding(2, 10)
    self.linear = nn.Linear(in_features=config.hidden_size + 10, out_features=n_classes)

  def forward(self,
              input,
              view_type="primary",
              final_multi_head_repr=None,
              input_mask=None,
              segment_ids=None,
              extra_args=None,
              **kwargs):
    # hard code
    predicate_indicator = extra_args["is_predicate_id"].long()
    predicate_ind_embed = self.predicate_indicator_embedding(predicate_indicator)
    logits = self.linear(torch.cat([input, predicate_ind_embed], dim=-1))
    return logits



import numpy as np
import torch
from typing import Optional
from relogic.logickit.base import utils


def print_2d_tensor(tensor):
  """ Print a 2D tensor """
  utils.log("lv, h >\t" + "\t".join(f"{x + 1}" for x in range(len(tensor))))
  for row in range(len(tensor)):
    if tensor.dtype != torch.long:
      utils.log(f"layer {row + 1}:\t" + "\t".join(f"{x:.5f}" for x in tensor[row].cpu().data))
    else:
      utils.log(f"layer {row + 1}:\t" + "\t".join(f"{x:d}" for x in tensor[row].cpu().data))

def entropy(p):
  """Compute the entropy of a probability distribution"""
  plogp = p * torch.log(p)
  plogp[p == 0] = 0
  return -plogp.sum(dim=-1)

def softmax(x):
  """Compute softmax values for each sets of scores in x."""
  e_x = np.exp(x - np.max(x))
  return e_x / e_x.sum(axis=-1)

def sigmoid(x):
  """Compute sigmoid values"""
  return np.exp(-np.logaddexp(0, -x))

def gen_position_indicator(span, length):
  indicator = [0] * length
  for idx, i in enumerate(range(span[0], -1, -1)):
    indicator[i] = -idx
  for idx, i in enumerate(range(span[1], length)):
    indicator[i] = idx + 1
  return indicator

def indicator_vector(index, length, default_label=0, indicator_label=1, head_index=None):
  vector = [default_label] * length
  if head_index is None:
    for idx in index:
      vector[idx] = indicator_label
  else:
    for idx in index:
      vector[head_index[idx]] = indicator_label
  return vector

def truncate_seq_pair(tokens_a, tokens_b, max_length):
  """Truncates a sequence pair in place to the maximum length."""

  # This is a simple heuristic which will always truncate the longer sequence
  # one token at a time. This makes more sense than truncating an equal percent
  # of tokens from each, since if one sequence is very short then each token
  # that's truncated likely contains more information than a longer sequence.
  while True:
    total_length = len(tokens_a) + len(tokens_b)
    if total_length <= max_length:
      break
    if len(tokens_a) > len(tokens_b):
      tokens_a.pop()
    else:
      tokens_b.pop()

def get_span_labels(sentence_tags, is_head=None, segment_id=None, inv_label_mapping=None, ignore_label=list([])):
  """Go from token-level labels to list of entities (start, end, class)."""
  if inv_label_mapping:
    sentence_tags = [inv_label_mapping[i] for i in sentence_tags]
  filtered_sentence_tag = []
  if is_head:
    # assert(len(sentence_tags) == len(is_head))

    for idx, (head, segment) in enumerate(zip(is_head, segment_id)):
      if (head == 1 or head == True) and (segment == 0 or segment == True):
        if sentence_tags[idx] != 'X':
          filtered_sentence_tag.append(sentence_tags[idx])
        else:
          filtered_sentence_tag.append("O")
  if filtered_sentence_tag:
    sentence_tags = filtered_sentence_tag
  span_labels = []
  last = 'O'
  start = -1
  for i, tag in enumerate(sentence_tags):
    items = (None, 'O') if tag == 'O' else tag.split('-', 1)
    pos, _ = items if len(items) == 2 else (items[0], None)
    if (pos == 'S' or pos == 'B' or tag == 'O') and last != 'O':
      span_labels.append((start, i - 1, None if len(last.split('-', 1)) != 2 else last.split('-', 1)[-1]))
    if pos == 'B' or pos == 'S' or last == 'O':
      start = i
    last = tag
  if sentence_tags[-1] != 'O':
    span_labels.append((start, len(sentence_tags) - 1,
                        None if len(last.split('-', 1)) != 2 else last.split('-', 1)[-1]))

  # This code has problem!
  # for item in span_labels:
  #   if item[2] in ignore_label:
  #     span_labels.remove(item)

  filtered_labels = []
  for item in span_labels:
    if item[2] not in ignore_label:
      filtered_labels.append(item)

  return set(filtered_labels), sentence_tags

def filter_head_prediction(sentence_tags, is_head):
  filtered_sentence_tag = []
  for idx, head in enumerate(is_head):
    if head == 1 or head == True:
      if sentence_tags[idx] != 'X':
        filtered_sentence_tag.append(sentence_tags[idx])
      else:
        filtered_sentence_tag.append("O")
  return filtered_sentence_tag

def create_tensor(features, attribute, dtype, device):

  try:
    return torch.tensor([getattr(f, attribute) for f in features], dtype=dtype).to(device)
  except Exception as e:
    if attribute not in create_tensor.attribute_warning:
      print("Exception in attribute {}".format(attribute))
      print(e)
      create_tensor.attribute_warning.add(attribute)
    return None

create_tensor.attribute_warning = set([])

def get_range_vector(size: int, device) -> torch.Tensor:
  """
  """
  return torch.arange(0, size, dtype=torch.long).to(device)

def flatten_and_batch_shift_indices(indices: torch.LongTensor,
                                    sequence_length: int) -> torch.Tensor:
  """``indices`` of size ``(batch_size, d_1, ..., d_n)`` indexes into dimension 2 of a target tensor,
  which has size ``(batch_size, sequence_length, embedding_size)``. This function returns a vector
  that correctly indexes into the flattened target. The sequence length of the target must be provided
  to compute the appropriate offset.

  Args:
    indices (torch.LongTensor):

  """
  if torch.max(indices) >= sequence_length or torch.min(indices) < 0:
    raise ValueError("All the elements should be in range (0, {}), but found ({}, {})".format(
      sequence_length - 1, torch.min(indices).item(), torch.max(indices).item()))
  offsets = get_range_vector(indices.size(0), indices.device) * sequence_length
  for _ in range(len(indices.size()) - 1):
    offsets = offsets.unsqueeze(1)

  # (batch_size, d_1, ..., d_n) + (batch_size, 1, ..., 1)
  offset_indices = indices + offsets

  # (batch_size * d_1 * ... * d_n)
  offset_indices = offset_indices.view(-1)
  return offset_indices


def batched_index_select(target: torch.Tensor,
                         indices: torch.LongTensor,
                         flattened_indices: Optional[torch.LongTensor] = None) -> torch.Tensor:
  """Select ``target`` of size ``(batch_size, sequence_length, embedding_size)`` with ``indices`` of
  size ``(batch_size, d_1, ***, d_n)``.

  Args:
    target (torch.Tensor): A 3 dimensional tensor of shape (batch_size, sequence_length, embedding_size).

  """
  if flattened_indices is None:
    flattened_indices = flatten_and_batch_shift_indices(indices, target.size(1))

  # Shape: (batch_size * sequence_length, embedding_size)
  flattened_target = target.view(-1, target.size(-1))

  # Shape: (batch_size * d_1 * ... * d_n, embedding_size)
  flattened_selected = flattened_target.index_select(0, flattened_indices)
  selected_shape = list(indices.size()) + [target.size(-1)]

  # Shape: (batch_size, d_1, ..., d_n, embedding_size)
  selected_targets = flattened_selected.view(*selected_shape)
  return selected_targets

def batched_index_select_tensor(target: torch.Tensor,
                                indices: torch.LongTensor,
                                flattened_indices: Optional[torch.LongTensor] = None) -> torch.Tensor:
  """Select ``target`` of size ``(batch_size, sequence_length, d2, d3, ... ,dn)`` with indices of size
  ``(batch, index_of_sequence)``.
  Args:
    target (torch.Tensor): High dimensional tensor, at least 3-D.
    indices (torch.LongTensor): A 2 dimensional tensor of shape (batch_size, sequence_length)
  """
  if flattened_indices is None:
    flattened_indices = flatten_and_batch_shift_indices(indices, target.size(1))

  flatten_target = target.view((target.size(0) * target.size(1),) + (target.size()[2:]))

  flattened_selected = flatten_target.index_select(0, flattened_indices)
  selected_shape = indices.size() + target.size()[2:]

  selected_targets = flattened_selected.view(*selected_shape)
  return selected_targets


def masked_softmax(vector: torch.Tensor,
                   mask: torch.Tensor,
                   dim: int = -1,
                   mask_fill_value: float = -1e32) -> torch.Tensor:
  """
  ``torch.nn.functional.softmax(vector)`` does not work if some elements of ``vector`` should be
  masked. This performs a softmax on just the non-masked positions of ``vector``. Passing ``None``
  in for the mask is also acceptable, which is just the regular softmax.

  """
  if mask is None:
    result = torch.softmax(vector, dim=dim)
  else:
    mask = mask.float()
    while mask.dim() < vector.dim():
      mask = mask.unsqueeze(1)
    masked_vector = vector.masked_fill((1 - mask).byte(), mask_fill_value)
    result = torch.softmax(masked_vector, dim=dim)
  return result

def masked_log_softmax(vector: torch.Tensor, mask: torch.Tensor, dim: int = -1) -> torch.Tensor:
  """
  ``torch.nn.functional.log_softmax(vector)`` does not work if some elements of ``vector`` should be
  masked.  This performs a log_softmax on just the non-masked portions of ``vector``.  Passing
  ``None`` in for the mask is also acceptable; you'll just get a regular log_softmax.
  ``vector`` can have an arbitrary number of dimensions; the only requirement is that ``mask`` is
  broadcastable to ``vector's`` shape.  If ``mask`` has fewer dimensions than ``vector``, we will
  unsqueeze on dimension 1 until they match.  If you need a different unsqueezing of your mask,
  do it yourself before passing the mask into this function.
  In the case that the input vector is completely masked, the return value of this function is
  arbitrary, but not ``nan``.  You should be masking the result of whatever computation comes out
  of this in that case, anyway, so the specific values returned shouldn't matter.  Also, the way
  that we deal with this case relies on having single-precision floats; mixing half-precision
  floats with fully-masked vectors will likely give you ``nans``.
  If your logits are all extremely negative (i.e., the max value in your logit vector is -50 or
  lower), the way we handle masking here could mess you up.  But if you've got logit values that
  extreme, you've got bigger problems than this.
  """
  if mask is not None:
    mask = mask.float()
    while mask.dim() < vector.dim():
      mask = mask.unsqueeze(1)
    # vector + mask.log() is an easy way to zero out masked elements in logspace, but it
    # results in nans when the whole vector is masked.  We need a very small value instead of a
    # zero in the mask for these cases.  log(1 + 1e-45) is still basically 0, so we can safely
    # just add 1e-45 before calling mask.log().  We use 1e-45 because 1e-46 is so small it
    # becomes 0 - this is just the smallest value we can actually use.
    vector = vector + (mask + 1e-45).log()
  return torch.nn.functional.log_softmax(vector, dim=dim)

def weighted_sum(matrix: torch.Tensor,
                 attention: torch.Tensor) -> torch.Tensor:
  """

  Args:
    matrix ():
    attention ():

  """
  if attention.dim() == 2 and matrix.dim() == 3:
    return attention.unsqueeze(1).bmm(matrix).squeeze(1)
  if attention.dim() == 3 and matrix.dim() == 3:
    return attention.bmm(matrix)
  if matrix.dim() - 1 < attention.dim():
    expanded_size = list(matrix.size())
    for i in range(attention.dim() - matrix.dim() + 1):
      matrix = matrix.unsqueeze(1)
      expanded_size.insert(i + 1, attention.size(i + 1))
    matrix = matrix.expand(*expanded_size)
  intermediate = attention.unsqueeze(-1).expand_as(matrix) * matrix
  return intermediate.sum(dim=-2)


def replace_masked_values(tensor: torch.Tensor, mask: torch.Tensor, replace_with: float) -> torch.Tensor:
  """
  """
  if tensor.dim() != mask.dim():
    raise ValueError("tensor.dim() {} != mask.dim() {}.".format(tensor.dim(), mask.dim()))
  return tensor.masked_fill((1-mask).byte(), replace_with)

def get_mask_from_sequence_lengths(sequence_lengths: torch.Tensor, max_length: int) -> torch.Tensor:
  """Generate mask from variable ``(batch_size,)`` which represents the sequence lengths of each
  batch element.

  Returns:
    torch.Tensor: ``(batch_size, max_length)``
  """
  ones = sequence_lengths.new_ones(sequence_lengths.size(0), max_length)
  range_tensor = ones.cumsum(dim=1)
  return (range_tensor <= sequence_lengths.unsqueeze(1)).long()

def get_device_of(tensor: torch.Tensor) -> int:
  """
  Returns the device of the tensor.
  """
  if not tensor.is_cuda:
    return -1
  else:
    return tensor.get_device()